//-----------------------------------------------------------------------
#include <iostream>

#include "CNetwork.h"
#include "CException.h"
#include "INetworkUser.h"
//-----------------------------------------------------------------------
#include <iostream>
#include <cstring> 
using namespace std;
//-----------------------------------------------------------------------
Uint16         CNetwork::s_iPort=6333;
Uint16         CNetwork::s_iClientUDPPort=6334;
SDL_mutex     *CNetwork::s_MutexFifo=NULL;
SDL_mutex     *CNetwork::s_MutexBuffer=NULL;
//-----------------------------------------------------------------------
CNetwork::CNetwork()
{
  m_User=NULL;
  m_bError=false;
  // initialize SDL_net
  if (SDLNet_Init() == -1)
  {
    print("CNetwork - SDLNet_Init: %s\n",SDLNet_GetError());
    m_bNetworkAvailable=false;
  }
  else
  {
    print("  SDLNet successfully initialized.");
    m_bNetworkAvailable=true;
  }
  m_bThreadShouldStop=false;
  // init log file 
  FILE *f=fopen("net.log","w");
  if (NULL != f)
    fclose(f);
  // noting started
  m_bServerStarted=false;
  m_bClientStarted=false;
  // buffer
  m_Buffer=NULL;
  m_iLenBuffer=0;
  // mutex
  // -> fifo
  if (s_MutexFifo == NULL)
    s_MutexFifo=SDL_CreateMutex();
  else
    throw CException("CNetwork: cannot create another CNetwork instance !");
  if (s_MutexFifo == NULL)
  {
    printLog("CNetwork - SDL_CreateMutex: %s\n",SDLNet_GetError());
    print("CNetwork: no fifo mutex, network may be unstable !");
  }
  // -> buffer
  if (s_MutexBuffer == NULL)
    s_MutexBuffer=SDL_CreateMutex();
  else
    throw CException("CNetwork: cannot create another CNetwork instance !");
  if (s_MutexBuffer == NULL)
  {
    printLog("CNetwork - SDL_CreateMutex: %s\n",SDLNet_GetError());
    print("CNetwork: no buffer mutex, network may be unstable !");
  }
  // ids
  m_iClientId=0;
  m_iNextClientId=1;
  // not ready
  m_bReady=false;
  // a killer message ;)
  cerr <<endl<< "Ghost Network Library v"<<CNETWORK_VER_MAJ<<"."<<CNETWORK_VER_MIN<<" initialized." << endl <<endl;;
}
//-----------------------------------------------------------------------
CNetwork::~CNetwork()
{
	//m_bThreadShouldStop=true;
	if (m_TCPThread != NULL)
		SDL_KillThread(m_TCPThread);
	if (m_UDPThread != NULL)    
		SDL_KillThread(m_UDPThread);
	if (m_TCPSocket != NULL)
		SDLNet_TCP_Close(m_TCPSocket);
	if (m_UDPSocket != NULL)    
		SDLNet_UDP_Close(m_UDPSocket);
	if (s_MutexFifo != NULL)
		SDL_DestroyMutex(s_MutexFifo);
	s_MutexFifo=NULL;
	if (s_MutexBuffer != NULL)
		SDL_DestroyMutex(s_MutexBuffer);
	s_MutexBuffer=NULL;
	SDLNet_Quit();
}
//-----------------------------------------------------------------------
CNetwork *CNetwork::getUniqueInstance()
{
  static CNetwork *net=NULL;

  if (net == NULL)
    net=new CNetwork();
  return (net);
}
//-----------------------------------------------------------------------
void CNetwork::print(const char *msg,...)
{
    va_list args;
    va_start(args, msg);
    
    vfprintf(stderr,msg,args);
    fprintf(stderr,"\n");
    
    va_end(args);
}
//-----------------------------------------------------------------------
void CNetwork::printLog(const char *msg,...)
{
	FILE *f;
	va_list args;
	va_start(args, msg);

	f=fopen("net.log","a");
	if (f == NULL)
	{
		fprintf(stderr,"[LOG] ");
		vfprintf(stderr,msg,args);
	}
	else
	{
		fprintf(f,"[LOG] ");
		vfprintf(f,msg,args);
		fclose(f);
	}
	va_end(args);
}
//-----------------------------------------------------------------------
void CNetwork::bindTo(INetworkUser *u)
{
	m_User=u;
}
//-----------------------------------------------------------------------
//***********************************************************************
//                             Process Message
//***********************************************************************
//-----------------------------------------------------------------------
void CNetwork::processMessages()
{
  network_msg r;

  if (!isEnabled())
    return;
  // -> lock
  lockFifo();
  // process messages
  for (list<network_msg>::iterator I=m_Fifo.begin();I!=m_Fifo.end();I++)
  {
    //std::cerr << "Message type " << (*I).m_Type << " data size " << (*I).m_Data.m_iSize << std::endl;
    switch ((*I).m_Type)
    {
    case NTWRK_MSG_SYS_MESSAGE: // [BOTH]
      if ((*I).m_Util == NTWRK_TO_SERVER)
	print("(server) ");
      if ((*I).m_ClientId == CNETWORK_SERVER_ID)
	print("From server: ");
      else
	print("From %d: ",(*I).m_ClientId);
      print("%s\n",(char *)(*I).m_Data.m_Raw);
      break;
    case NTWRK_MSG_CONNECTION:  // [SERVER] client connection
      print("NTWRK_MSG_CONNECTION\n");
      break;
    case NTWRK_MSG_DISCONNECTION: // [SERVER] client disconnection
      {
	client_iterator C;
	print("(server) NTWRK_MSG_DISCONNECTION\n");
	C=m_Clients.begin();
	while (C!=m_Clients.end())
	{
	  if ((long int)((*C).second.m_Socket) == (*I).m_Util)
	  {
	    network_msg m;
	    messageFromData(NTWRK_MSG_CLIENT_DISCONNECTION,
			    CNETWORK_SERVER_ID,
			    NULL,0,&m);
	    m.m_Util=(*C).second.m_Id;
	    print("(server) Erasing client %d\n",m.m_Util);
	    m_Clients.erase(m.m_Util);
	    for (client_iterator D=m_Clients.begin();
		 D!=m_Clients.end();D++)
	      secureSend((*D).second.m_Socket,&m);
	    break;
	  }
	  C++;
	}
	break;
      }
    case NTWRK_MSG_CLIENT_DISCONNECTION: // [CLIENT] client disconnection
      print("NTWRK_MSG_CLIENT_DISCONNECTION\n");
      print("Erasing client %d\n",(*I).m_Util);
      m_Clients.erase((*I).m_Util);
      // warn network user
      if (NULL != m_User)
      {
	m_User->onClientDisconnection((*I).m_Util);
      }
      break;
    case NTWRK_MSG_ACK:
      print("NTWRK_MSG_ACK\n");
      break;
    case NTWRK_MSG_NACK:
      print("NTWRK_MSG_NACK\n");
      break;
    case NTWRK_MSG_WELCOME: // [CLIENT] client id
      {
	msg_welcome *w;
	msg_udp_port udp;

	print("NTWRK_MSG_WELCOME\n");
	// reading welcome data
	w=reinterpret_cast<msg_welcome *>((*I).m_Data.m_Raw);
	m_iClientId=w->m_Id;
	print("Client Id is %d\n",m_iClientId);

	// compare the mobinet version with the server one
	unsigned char serverVerMaj = w->m_verMaj;
	unsigned char serverVerMin = w->m_verMin;
	unsigned char serverVerRev = w->m_verRev;
	print("client version : %i.%i.%i ", m_ui8VerMaj, m_ui8VerMin, m_ui8VerRev);
	print("server version : %i.%i.%i \n", serverVerMaj, serverVerMin, serverVerRev);
	if (serverVerMaj==1 &&(serverVerMaj != this->m_ui8VerMaj || serverVerMin != this->m_ui8VerMin)){
		print("connection Error : client mobinet version (%i.%i) is different from server one (%i.%i).",
			m_ui8VerMaj, m_ui8VerMin, serverVerMaj, serverVerMin);
		throw CException("version incompatibility");
	}

	if ( serverVerRev != this->m_ui8VerRev){
		print("WARNING : client mobinet revision (%i.%i.%i) is quite different from server one (%i.%i.%i), you may have some troubles...",
			m_ui8VerMaj, m_ui8VerMin, m_ui8VerRev, serverVerMaj, serverVerMin, serverVerRev);
	}

	// sending UDP port
	udp.m_Port=SDL_SwapBE16(s_iClientUDPPort);
	udp.m_Id=m_iClientId;
	udp.m_verMaj = this->m_ui8VerMaj;
	udp.m_verMin = this->m_ui8VerMin;
	udp.m_verRev = this->m_ui8VerRev;
    messageFromData(NTWRK_MSG_UDP_PORT,m_iClientId,
			(unsigned char *)&udp,sizeof(udp),&r);
	secureSend(m_TCPSocket,&r);
	break;
      }
    case NTWRK_MSG_UDP_PORT: // [SERVER] udp info on client
      {
	msg_udp_port *udp;
	IPaddress     ip;
	Uint32        a;
	Uint16        p;

	print("NTWRK_MSG_UDP_PORT\n");
	// reading port
	udp=reinterpret_cast<msg_udp_port *>((*I).m_Data.m_Raw);
	ip.host=(*I).m_Ip.host;
	ip.port=udp->m_Port;

	//compare the client mobinet version with the server one
	unsigned char clientVerMaj = udp->m_verMaj;
	unsigned char clientVerMin = udp->m_verMin;
	unsigned char clientVerRev = udp->m_verRev;
	print("server version : %i.%i.%i ", m_ui8VerMaj, m_ui8VerMin, m_ui8VerRev);
	print("client version : %i.%i.%i \n", udp->m_verMaj, udp->m_verMin, udp->m_verRev);
	if( clientVerMaj==1 && (m_ui8VerMaj != udp->m_verMaj || m_ui8VerMin != udp->m_verMin ) ){
		print("Client connection Error : client mobinet version (%i.%i) is different of server one (%i.%i), must be %i.x.",
			clientVerMaj, clientVerMin, m_ui8VerMaj, m_ui8VerMin, m_ui8VerMaj);
		
		break;
	}
	
	// adding client
	network_client& cli=m_Clients[udp->m_Id];
	cli.m_Socket=(*I).m_Socket;
	cli.m_UDPAddress=ip;
	cli.m_Id=udp->m_Id;
	a=SDL_SwapBE32(cli.m_UDPAddress.host);
	p=SDL_SwapBE16(cli.m_UDPAddress.port);
	print("(Server) Adding client id=%d ip=%d.%d.%d.%d port=%d\n",
	      cli.m_Id,a>>24,(a>>16)&0xff,(a>>8)&0xff,a&0xff,p);
	printLog("(Server) Adding client id=%d ip=%d.%d.%d.%d port=%d\n",
		 cli.m_Id,a>>24,(a>>16)&0xff,(a>>8)&0xff,a&0xff,p);
	// warn network user
	if (NULL != m_User)
	  m_User->onClientConnection(cli.m_Id);
	// everybody should know everybody else
	updateNetwork(cli);
	break;
      }
    case NTWRK_MSG_ADD_CLIENTS: // [CLIENT] add other client info
      {
	msg_client *c;
	int         i,n;
	Uint32      a;
	Uint16      p;

	n=(*I).m_Data.m_iSize/sizeof(msg_client);
	print("%d clients received\n",n);
	for (i=0;i<n;i++)
	{
	  c=reinterpret_cast<msg_client *>((*I).m_Data.m_Raw + 
					   i*sizeof(msg_client));
	  if (m_Clients.find(c->m_Id) != m_Clients.end())
	    throw CException("CNetwork: duplicate client id (%d)",
			     c->m_Id);
	  if (c->m_Id == 0) // client on same computer as server
	    c->m_Ip=m_IpServer;
	  network_client& cli=m_Clients[c->m_Id];
	  cli.m_Socket=NULL;
	  cli.m_UDPAddress=c->m_Ip;
	  cli.m_Id=c->m_Id;
	  a=SDL_SwapBE32(cli.m_UDPAddress.host);
	  p=SDL_SwapBE16(cli.m_UDPAddress.port);
	  print("Adding client id=%d ip=%d.%d.%d.%d port=%d\n",
		c->m_Id,a>>24,(a>>16)&0xff,(a>>8)&0xff,a&0xff,p);
	  printLog("Adding client id=%d ip=%d.%d.%d.%d port=%d\n",
		   c->m_Id,a>>24,(a>>16)&0xff,(a>>8)&0xff,a&0xff,p);
	  // warn network user
	  if (NULL != m_User)
	    m_User->onClientConnection(c->m_Id);
	}
	// here we can be sure that our client is ready to work
	m_bReady=true;
      }
      break;
    case NTWRK_MSG_RELAY:
      {
	network_msg m;

	// print("NTWRK_MSG_RELAY\n");
	serverSend(NTWRK_SECURE,(*I).m_Util,
		   extractRelayMessage(&(*I),&m),true);
	freeMessageData(m);
	break;
      }
    default:
      if (m_User != NULL)
      {
	if ((*I).m_Util == NTWRK_TO_SERVER)
	{
	  if (!isServer())
	    throw CException("CNetwork::processMessages - I'm not a server !");
	  //std::cerr << "CNetwork::processMessages: user server" << endl;
	  m_User->processServerMessage(*I);
	}
	else
	{
	  //std::cerr << "CNetwork::processMessages: user client" << endl;
	  m_User->processClientMessage(*I);
	}
      }
      else
	print("CNetwork::processMessages: no user, discarding unknwon message.\n");
      break;
    }
    freeMessageData((*I));
  }
  m_Fifo.clear();
  // -> unlock
  unlockFifo();
}
//-----------------------------------------------------------------------
void CNetwork::updateNetwork(const network_client& cli)
{
  int            sz;
  network_msg    m;
  msg_client     c;
  unsigned char *buf,*ptr;

  sz=(m_Clients.size()-1)*sizeof(msg_client);
  buf=new unsigned char [sz];
  ptr=buf;
  for (client_iterator I=m_Clients.begin();I!=m_Clients.end();I++)
  {
    if (cli.m_UDPAddress.host == (*I).second.m_UDPAddress.host 
	&& cli.m_UDPAddress.port == (*I).second.m_UDPAddress.port)
      continue;
    // for new client
    c.m_Ip=(*I).second.m_UDPAddress;
    c.m_Id=(*I).second.m_Id;
    memcpy(ptr,&c,sizeof(c));
    ptr+=sizeof(c);
    // for old client
    if ((*I).second.m_Socket == NULL) // do not send information to 
      // server client (already 
      // build by NTWRK_UDP_PORT)
      continue;
    c.m_Ip=cli.m_UDPAddress;
    c.m_Id=cli.m_Id;
    secureSend((*I).second.m_Socket,
	       messageFromData(NTWRK_MSG_ADD_CLIENTS,CNETWORK_SERVER_ID,
			       (unsigned char *)&c,
			       sizeof(c),&m));
  }
  secureSend(cli.m_Socket,messageFromData(NTWRK_MSG_ADD_CLIENTS,
					  CNETWORK_SERVER_ID,
					  buf,sz,&m));
  delete [](buf);
}
//-----------------------------------------------------------------------
//***********************************************************************
//                          Events from threads
//***********************************************************************
//-----------------------------------------------------------------------
void CNetwork::onClientConnect(const TCPsocket& sock)
{
	network_msg m;
	msg_welcome w;

	w.m_Id=m_iNextClientId++;
	w.m_verMaj = this->m_ui8VerMaj;	
	w.m_verMin = this->m_ui8VerMin;
	w.m_verRev = this->m_ui8VerRev;
	//print("Sending welcome to new client (id %d)\n",w.m_Id);
	secureSend(sock,messageFromData(NTWRK_MSG_WELCOME,CNETWORK_SERVER_ID,
		(unsigned char *)&w,sizeof(w),&m));
}
//-----------------------------------------------------------------------
void CNetwork::onClientDisconnect(const TCPsocket& sock)
{
	network_msg m;

	// warn himself (cannot directly acces m_Clients: no mutex on it
	messageFromData(NTWRK_MSG_DISCONNECTION,CNETWORK_SERVER_ID,
		NULL,0,&m);
	m.m_Util=(long int)sock;
	storeMessage(&m);
}
//-----------------------------------------------------------------------
//***********************************************************************
//                               Mutex
//***********************************************************************
//-----------------------------------------------------------------------
inline void CNetwork::lockFifo()
{
	int startTime;

	if (s_MutexFifo == NULL)
		return;
	// -> lock
	startTime=SDL_GetTicks();
	while (SDL_mutexP(s_MutexFifo) == -1)
	{
		if (SDL_GetTicks() - startTime > CNETWORK_MUTEX_TMOUT)
			throw CException("CNetwork: mutex lock timeout !");
	}
}
//-----------------------------------------------------------------------
inline void CNetwork::unlockFifo()
{
	if (s_MutexFifo == NULL)
		return;
	// -> unlock
	if(SDL_mutexV(s_MutexFifo) == -1)
		throw CException("CNetwork: mutex unlock fault !");
}
//-----------------------------------------------------------------------
inline void CNetwork::lockBuffer()
{
	int startTime;

	if (s_MutexBuffer == NULL)
		return;
	// -> lock
	startTime=SDL_GetTicks();
	while (SDL_mutexP(s_MutexBuffer) == -1)
	{
		if (SDL_GetTicks() - startTime > CNETWORK_MUTEX_TMOUT)
			throw CException("CNetwork: mutex lock timeout !");
	}
}
//-----------------------------------------------------------------------
inline void CNetwork::unlockBuffer()
{
	if (s_MutexBuffer == NULL)
		return;
	// -> unlock
	if(SDL_mutexV(s_MutexBuffer) == -1)
		throw CException("CNetwork: mutex unlock fault !");
}
//-----------------------------------------------------------------------
//***********************************************************************
//                             Message handling
//***********************************************************************
//-----------------------------------------------------------------------
void CNetwork::allocBuffer(int needed,unsigned char *& buf,int& bufl)
{
  if (buf == NULL)
  {
    buf=new unsigned char[needed];
    bufl=needed;
  }
  else
  {
    if (needed > bufl)
    {
      delete [](buf);
      buf=new unsigned char[needed];
      bufl=needed;
    }
  }
}
//-----------------------------------------------------------------------
int CNetwork::prepareMessage(network_msg *m,
							 unsigned char *& buf,
							 int& buflen)
{
	int    size;
	Uint32 type;
	Uint32 util;
	Uint8  id;

	size=sizeof(m->m_Type)
		+sizeof(m->m_Util)
		+sizeof(m->m_ClientId)
		+m->m_Data.m_iSize;
	allocBuffer(size,buf,buflen);
	// type
	type=m->m_Type;
	//  type=SDL_SwapBE32(type);
	memcpy(buf,&type,sizeof(type));
	// util
	util=m->m_Util;
	//  util=SDL_SwapBE32(util);
	memcpy(buf+sizeof(type),&util,sizeof(util));
	// id
	id=m->m_ClientId;
	memcpy(buf+sizeof(type)+sizeof(util),&id,sizeof(id));  
	// data
	if (m->m_Data.m_iSize > 0)
		memcpy(buf+sizeof(type)+sizeof(util)+sizeof(id),
		m->m_Data.m_Raw,m->m_Data.m_iSize);
	return (size);
}
//-----------------------------------------------------------------------
void CNetwork::extractMessage(network_msg& msg,unsigned char *buf,int len)
{
  Uint32 type;
  Uint32 util;
  Uint8  id;

  // extract type
  memcpy(&type,buf,sizeof(type));
  //  type=SDL_SwapBE32(type);
  // extract util
  memcpy(&util,buf+sizeof(type),sizeof(util));
  //  util=SDL_SwapBE32(util);
  // extract id
  memcpy(&id,buf+sizeof(type)+sizeof(util),sizeof(id));
  // -> store message
  msg.m_Type=type;
  msg.m_Util=util;
  msg.m_ClientId=id;
  msg.m_Ip.host=0;
  msg.m_Ip.port=0;
  msg.m_Socket=NULL;
  msg.m_bLocal=false;
  msg.m_Data.m_iSize=len-sizeof(type)-sizeof(util)-sizeof(id);
  if (msg.m_Data.m_iSize > 0)
  {
    msg.m_Data.m_Raw=new unsigned char [msg.m_Data.m_iSize];
    memcpy(msg.m_Data.m_Raw,
	   buf+sizeof(type)+sizeof(util)+sizeof(id),
	   msg.m_Data.m_iSize);
  }
  else
  {
    msg.m_Data.m_Raw=NULL;
  }
  /*
    std::cerr << "Received message type=" << type 
    << " id=" << (int)id 
    << " util=" << util 
    << endl;
  */
}
//-----------------------------------------------------------------------
void CNetwork::extractMessage(const TCPsocket s,const IPaddress *i,
			      unsigned char *buf,int len)
{
  static network_msg empty;

  // -> lock
  lockFifo();
  // -> extract
  m_Fifo.push_back(empty);        // empty is needed for Visual C++ STL compatibility
  network_msg &m=m_Fifo.back();
  extractMessage(m,buf,len);
  // -> fill
  if (s == NULL && i == NULL)
  {
    m.m_Ip=m_IpServer;
    m.m_bFromServer=true;
  }
  else if (s != NULL)
  {
    IPaddress *ip;

    m.m_Socket=s;
    ip=SDLNet_TCP_GetPeerAddress(s);
    if (ip)
      m.m_Ip=*ip;
  }
  else
  {
    m.m_Ip=*i;
  }
  // -> unlock
  unlockFifo();
}
//-----------------------------------------------------------------------
void CNetwork::storeMessage(const network_msg *m)
{
  static network_msg empty;
  bool from_server=(m->m_ClientId==CNETWORK_SERVER_ID);

  // add message to fifo with mutex !
  // -> lock
  lockFifo();
  // -> store a copy of the message
  m_Fifo.push_back(empty);
  network_msg &msg=m_Fifo.back();
  memcpy(&msg,m,sizeof(network_msg));
  msg.m_Data.m_Raw=new unsigned char [msg.m_Data.m_iSize];
  memcpy(msg.m_Data.m_Raw,m->m_Data.m_Raw,msg.m_Data.m_iSize);
  msg.m_Ip.host=0;
  msg.m_Ip.port=0;
  msg.m_bLocal=true;
  msg.m_bFromServer=from_server;
  msg.m_ClientId=m->m_ClientId;
  // -> unlock
  unlockFifo();
}
//-----------------------------------------------------------------------
void CNetwork::freeMessageData(network_msg& m)
{
  if (m.m_Data.m_Raw != NULL)
  {
    delete [](m.m_Data.m_Raw);
    m.m_Data.m_Raw=NULL;
  }
  m.m_Data.m_iSize=0;
}
//-----------------------------------------------------------------------
network_msg *CNetwork::messageFromData(int t,Uint8 id,
									   unsigned char *d,int l,
									   network_msg *tmp)
{
  memset(tmp,0,sizeof(network_msg));
  tmp->m_Type=t;
  tmp->m_Data.m_iSize=l;
  tmp->m_Data.m_Raw=d;
  tmp->m_ClientId=id;
  return (tmp);
}
//-----------------------------------------------------------------------
network_msg *CNetwork::clientMessageFromData(int t,
											 unsigned char *d,int l,
											 network_msg *tmp)
{
  return (messageFromData(t,clientId(),d,l,tmp));
}
//-----------------------------------------------------------------------
network_msg *CNetwork::serverMessageFromData(int t,
											 unsigned char *d,int l,
											 network_msg *tmp)
{
  if (!isServer())
    throw CException("CNetwork::serverMessageFromData - I'm not server !");
  return (messageFromData(t,CNETWORK_SERVER_ID,d,l,tmp));
}
//-----------------------------------------------------------------------
network_msg *CNetwork::buildRelayMessage(network_msg *src,int dest,
										 network_msg *dst)
{
  static unsigned char *pBuf=NULL;
  static unsigned char *sBuf=NULL; // no need of mutex, only one thread
  static int            lBuf=0;
  int                   len;

  memset(dst,0,sizeof(network_msg));
  dst->m_Type=NTWRK_MSG_RELAY;
  dst->m_Util=dest;
  len=prepareMessage(src,sBuf,lBuf);
  dst->m_Data.m_iSize=len;
  dst->m_Data.m_Raw=sBuf;
  dst->m_ClientId=clientId();
  return (dst);  
}
//-----------------------------------------------------------------------
network_msg *CNetwork::extractRelayMessage(network_msg *src,
					   network_msg *dst)
{
  memset(dst,0,sizeof(network_msg));
  extractMessage(*dst,src->m_Data.m_Raw,src->m_Data.m_iSize);
  return (dst);
}
//-----------------------------------------------------------------------
//***********************************************************************
//                             Send methods
//***********************************************************************
//-----------------------------------------------------------------------
void CNetwork::serverSend(int method,int dest,network_msg *m,bool relay)
{
  int size;

  if (!isEnabled())
    return;
  // check if we are server
  if (!m_bServerStarted)
    throw CException("CNetwork::serverSend - I am not server !");
  lockBuffer();
  // message sender
  if (!relay)
    m->m_ClientId=CNETWORK_SERVER_ID;
  // translation of dest for simplicity
  if (dest == NTWRK_TO_OTHERS && !relay)
    dest=NTWRK_TO_ALL;
  // prepare message
  size=prepareMessage(m,m_Buffer,m_iLenBuffer);
  // send to dest
  switch (dest)
  {
  case NTWRK_TO_SERVER:
    throw CException("CNetwork::serverSend - server talking to himself ?!?");
    break;
  case NTWRK_TO_ALL:
  case NTWRK_TO_OTHERS:

    {
      for (client_iterator I=m_Clients.begin();
	   I!=m_Clients.end();I++)
      {
	if (dest == NTWRK_TO_OTHERS && (*I).first == m->m_ClientId)
	  continue;
	if ((*I).first == 0) // server client
	  storeMessage(m);
	else
	{
	  if (method == NTWRK_SECURE)
	    sendTCPData((*I).second.m_Socket,m_Buffer,size);
	  else
	    sendUDPData(m_UDPSocket,(*I).second.m_UDPAddress,
			m_Buffer,size);
	}
      }

	}
  break;
  default:
    {
      client_iterator C=m_Clients.find(dest);
      if (C == m_Clients.end())
	throw CException("serverSend - invalid dest (%d) !",dest);
      if (dest == 0) // server client
	storeMessage(m);
      else
      {
	if (method == NTWRK_SECURE)
	  sendTCPData((*C).second.m_Socket,m_Buffer,size);
	else
	  sendUDPData(m_UDPSocket,(*C).second.m_UDPAddress,
		      m_Buffer,size);
      }
    }
  }
  unlockBuffer();
}
//-----------------------------------------------------------------------
void CNetwork::clientSend(int method,int dest,network_msg *m)
{
  network_msg relay;
  int         size;

  if (!isEnabled())
    return;
  lockBuffer();
  // message sender
  m->m_ClientId=clientId();
  // prepare message
  if (method == NTWRK_SECURE 
      && dest != NTWRK_TO_SERVER 
      && dest != clientId())
    m=buildRelayMessage(m,dest,&relay);
  else if (dest == NTWRK_TO_SERVER)
    m->m_Util=NTWRK_TO_SERVER;
  size=prepareMessage(m,m_Buffer,m_iLenBuffer);
  // send to dest
  switch (dest)
  {
  case NTWRK_TO_ALL:
  case NTWRK_TO_OTHERS:
    if (method == NTWRK_SECURE) // relaying
    {
      if (clientId() == 0)
	storeMessage(m);
      else
	sendTCPData(m_TCPSocket,m_Buffer,size);
    }
    else
    {
      for (client_iterator I=m_Clients.begin();
	   I!=m_Clients.end();I++)
      {
	// server client should not send to himself (see below for TO_ALL)
	if ((*I).first == 0 && clientId() == 0) 
	  continue;
	sendUDPData(m_UDPSocket,(*I).second.m_UDPAddress,
		    m_Buffer,size);
      }
      // should I receive this ?
      if (dest == NTWRK_TO_ALL)
	storeMessage(m);
    }
    break;
  case NTWRK_TO_SERVER:
    if (method == NTWRK_SECURE)
    {
      if (clientId() == 0)
	storeMessage(m);
      else
	sendTCPData(m_TCPSocket,m_Buffer,size);
    }
    else
      sendUDPData(m_UDPSocket,m_IpServer,m_Buffer,size);
    break;
  default:
    {
      if (dest == clientId())
      {
	storeMessage(m);
      }
      else
      {
	client_iterator C=m_Clients.find(dest);
	if (C == m_Clients.end())
	  throw CException("serverSend - invalid dest (%d) !",dest);
	if (method == NTWRK_SECURE) // relaying
	{
	  if (clientId() == 0)
	    storeMessage(m);
	  else
	    sendTCPData(m_TCPSocket,m_Buffer,size);
	}
	else
	  sendUDPData(m_UDPSocket,(*C).second.m_UDPAddress,
		      m_Buffer,size);
      }
    }
  }
  unlockBuffer();
}
//-----------------------------------------------------------------------
void CNetwork::secureSend(TCPsocket sock,network_msg *m)
{
  int size;

  if (!isEnabled())
    return;
  lockBuffer();
  size=prepareMessage(m,m_Buffer,m_iLenBuffer);
  if (sock == NULL)
    storeMessage(m);
  else
    sendTCPData(sock,m_Buffer,size);
  unlockBuffer();
}
//-----------------------------------------------------------------------
void CNetwork::fastSend(const IPaddress& ip,network_msg *m)
{
  int size;

  if (!isEnabled())
    return;
  lockBuffer();
  size=prepareMessage(m,m_Buffer,m_iLenBuffer);
  sendUDPData(m_UDPSocket,ip,m_Buffer,size);
  unlockBuffer();
}
//-----------------------------------------------------------------------
//***********************************************************************
//                           Start functions
//***********************************************************************
//-----------------------------------------------------------------------
void CNetwork::startServer(unsigned char verMaj, unsigned char verMin, unsigned char verRev)
{
  if (!m_bNetworkAvailable)
    return;

  // set client version
  m_ui8VerMaj = verMaj;
  m_ui8VerMin = verMin;
  m_ui8VerRev = verRev;

  // Resolve the argument into an IPaddress type
  if (SDLNet_ResolveHost(&m_IpServer,NULL,s_iPort)==-1)
  {
    throw CException("CNetwork::startServer error - SDLNet_ResolveHost: %s\n",SDLNet_GetError());
  }

  // output the port number
  print("Server on port       : %d\n",s_iPort);

  // open the server TCP socket
  m_TCPSocket=SDLNet_TCP_Open(&m_IpServer);
  if(!m_TCPSocket)
  {
    throw CException("SDLNet_TCP_Open: Impossible d'ouvrir la connexion TCP\n -> verifiez que vous etes bien connecte au reseau\n -> le port %d est peut etre occupe ?\n -> votre application est peut-etre deja lancee ?\n\n",s_iPort);
    //throw CException("SDLNet_TCP_Open: %s\n",SDLNet_GetError());
  }

  // open the server UDP socket
  m_UDPSocket=SDLNet_UDP_Open(s_iPort);
  if(!m_UDPSocket)
  {
    throw CException("SDLNet_UDP_Open: %s\n",SDLNet_GetError());
  }

  printLog("Server sockets started.\n");

  // creating corresponding client
  network_client &cli=m_Clients[0];
  cli.m_Id=0;
  cli.m_Socket=NULL;
  cli.m_UDPAddress.host=0;
  cli.m_UDPAddress.port=SDL_SwapBE16(s_iPort);

  // start threads
  startServerThreads();

  m_bServerStarted=true;

  // the server is immediately ready
  m_bReady=true;
}
//-----------------------------------------------------------------------
void CNetwork::startClient(char *host_name, unsigned char verMaj, unsigned char verMin, unsigned char verRev)
{
  Uint32    ipaddr;
  const char     *host;

  if (!m_bNetworkAvailable)
    return;

  // set client version
  m_ui8VerMaj = verMaj;
  m_ui8VerMin = verMin;
  m_ui8VerRev = verRev;

  // Resolve the argument into an IPaddress type
  if (SDLNet_ResolveHost(&m_IpServer,host_name,s_iPort)==-1)
  {
    m_bNetworkAvailable=false;
    throw CException("CNetwork::startClient error - SDLNet_ResolveHost: %s\n",SDLNet_GetError());
  }
  // perform a byte endianess correction for the next printf
  ipaddr=SDL_SwapBE32(m_IpServer.host);

  // output the IP address nicely
  print("Server IP Address : %d.%d.%d.%d\n",
	ipaddr>>24,
	(ipaddr>>16)&0xff,
	(ipaddr>>8)&0xff,
	ipaddr&0xff);

  // resolve the hostname for the IPaddress
  host=SDLNet_ResolveIP(&m_IpServer);
  // print out the hostname we got
  if(host)
    print("Hostname   : %s\n",host);
  else
    print("Hostname   : N/A\n");
  // output the port number
  print("TCP Port       : %d\n",s_iPort);

  // open the client TCP socket
  m_TCPSocket=SDLNet_TCP_Open(&m_IpServer);
  if(!m_TCPSocket)
  {
    throw CException("SDLNet_TCP_Open: %s\n",SDLNet_GetError());
  }

  // open the client UDP socket
  m_UDPSocket=SDLNet_UDP_Open(s_iClientUDPPort);
  if(!m_UDPSocket)
  {
    throw CException("SDLNet_UDP_Open: %s\n",SDLNet_GetError());
  }
  print("UDP Port       : %d\n",s_iClientUDPPort);

  printLog("Client sockets started.\n");

  // start threads
  startClientThreads();    

  m_bClientStarted=true;
}
//-----------------------------------------------------------------------
bool CNetwork::isServer() 

{

	return (m_bServerStarted);

}

//-----------------------------------------------------------------------

bool CNetwork::isClient() 

{

	return (m_bClientStarted);

}

//-----------------------------------------------------------------------

bool CNetwork::isEnabled()

{

	return (m_bNetworkAvailable && (m_bServerStarted || m_bClientStarted));

}

//-----------------------------------------------------------------------

bool CNetwork::isReady()  
{

	return (m_bReady);

}

//-----------------------------------------------------------------------

//***********************************************************************
//                             Threads
//***********************************************************************
//-----------------------------------------------------------------------
int CNetwork_TCP_server_thread(void *data)
{
 SDLNet_SocketSet          sset = NULL;
 unsigned char            *buf  = NULL;
 CNetwork  *net  = (CNetwork *)data;
 int                       numready;
 list<TCPsocket>           sockets;
 list<TCPsocket>::iterator C;

 try {
   net->printLog("NETWORK_TCP_server_thread started.\n");      while ( !net->threadShouldStop() ) {
     // prepare socket set
     if (sset) {
       SDLNet_FreeSocketSet(sset);
     }
     sset = SDLNet_AllocSocketSet(sockets.size()+1);
     if (sset == NULL) {
       throw CException("SDLNet_AllocSocketSet: %s\n", SDLNet_GetError());
     }
     if (SDLNet_TCP_AddSocket(sset,net->m_TCPSocket) == -1) {
       throw CException("SDLNet_TCP_AddSocket: %s\n",SDLNet_GetError());
     }
     for (list<TCPsocket>::iterator I=sockets.begin() ;    I!=sockets.end() ; I++)    {
       if (SDLNet_TCP_AddSocket(sset,(*I)) == -1) {
         throw CException("SDLNet_TCP_AddSocket: %s\n",SDLNet_GetError());
       }
     }
     // listen (connection + incoming data)
     numready = SDLNet_CheckSockets(sset, (unsigned int)-1);
     // error ?
     if (numready == -1)    {
       throw CException("SDLNet_CheckSockets: %s\n",SDLNet_GetError());
     }
     // nobody ... strange
     if (numready == 0) {
       continue;
     }
     // is the server receiving a new client ?
     if (SDLNet_SocketReady(net->m_TCPSocket)) {
       TCPsocket sock;
       numready--;
       // accept connection
       sock = SDLNet_TCP_Accept(net->m_TCPSocket);
       if (sock != NULL) {
         IPaddress *remoteip = NULL;
         Uint32     ipaddr   = 0;
         Uint16     port     = 0;
         remoteip            = SDLNet_TCP_GetPeerAddress(sock);
         if (!remoteip)    {
           net->printLog("[WARNING] SDLNet_TCP_GetPeerAddress: %s\n",SDLNet_GetError());
         }    else {
           // print out the clients IP and port number
           ipaddr = SDL_SwapBE32(remoteip->host);
           port   = SDL_SwapBE16(remoteip->port);
           net->printLog("NETWORK_TCP_thread: Connection from %d.%d.%d.%d port %hu\n",
             (ipaddr>>24),
             (ipaddr>>16)&0xff,
             (ipaddr>> 8)&0xff,
             ipaddr&0xff,port);
         }
         // reading data sent
         if (!CNetwork::getTCPData(sock,&buf))    {
           char *errstr = SDLNet_GetError();
           net->printLog("[WARNING] (TCP thread - on connection) getTCPData: %s\n",strlen(errstr)?errstr:"disconnected");
           SDLNet_TCP_Close(sock);
         }    else {
           if (!strcmp((char *)buf,CNETWORK_ID)) {
             // add client
             sockets.push_back(sock);
             // connection event
             net->onClientConnect(sock);
           }    else {
             net->print(0,"[WARNING] Refusing client (bad handshake)\n");
             CNetwork::sendTCPData( sock, (unsigned char *)CNETWORK_REFUSE_MSG,
										strlen(CNETWORK_REFUSE_MSG)+1);
										SDLNet_TCP_Close(sock);
           }
         }
       }
     }
     // check if clients sent data (they are supressed if disconnected)
     C = sockets.begin();
     while ( C != sockets.end() ) {
       bool incC = true;
       if (SDLNet_SocketReady(*C))    {
         int res;
         numready--;
         // reading data sent
         res = CNetwork::getTCPData((*C),&buf);
         if (res <= 0) {
           if (res < 0) {
             char *errstr = SDLNet_GetError();
             net->printLog("[WARNING] (TCP thread) getTCPData: %s\n",strlen(errstr)?errstr:"disconnected");
           }
           // remove client
           TCPsocket tmp = (*C);
           C    = sockets.erase(C);
           incC = false;
           // disconnection event
           net->onClientDisconnect(tmp);
         } else {
           // message
           net->extractMessage((*C),NULL,buf,res);
         }
       }
       if (incC) {
         C++;
       }
     }
     if (numready != 0) {
       cerr << "WARNING: numready > 0 (" << numready << ")" << endl;
     }
   }
   // close connection
   for (list<TCPsocket>::iterator I=sockets.begin() ; I!=sockets.end() ; I++) {
     SDLNet_TCP_Close(*I);
   }
   // kill the thread
   SDL_KillThread(net->m_TCPThread);
   net->m_TCPThread = NULL;
 } catch (CException &f) {
   cerr << "FATAL ERROR: " << f.getMsg() << endl;
   throw f;
 }
 return (0);
}


//-----------------------------------------------------------------------
int CNetwork_TCP_client_thread(void *data)
{
  SDLNet_SocketSet sset=NULL;
  int              numready;
  unsigned char   *buf=NULL;
  CNetwork        *net=(CNetwork *)data;

  net->printLog("CNetwork_TCP_client_thread started.\n");	
  // prepare socket set
  if (sset)
    SDLNet_FreeSocketSet(sset);
  sset=SDLNet_AllocSocketSet(1);
  if (sset == NULL)
    throw CException("SDLNet_AllocSocketSet: %s\n", SDLNet_GetError());
  if (SDLNet_TCP_AddSocket(sset,net->m_TCPSocket) == -1)
    throw CException("SDLNet_TCP_AddSocket: %s\n",SDLNet_GetError());
  // send connection id to server
  CNetwork::sendTCPData(net->m_TCPSocket,
			(unsigned char *)CNETWORK_ID,
			strlen(CNETWORK_ID)+1);
  // go for loop ;)
  while (!net->threadShouldStop())
  {
    // listen (incoming data)
    numready=SDLNet_CheckSockets(sset, (unsigned int)-1);
    // error ?
    if (numready == -1)
      throw CException("SDLNet_CheckSockets: %s\n",SDLNet_GetError());
    // nobody ... strange
    if (numready == 0)
      continue;
    if (SDLNet_SocketReady(net->m_TCPSocket))
    {
      int res;

      numready--;
      // reading data sent
      res=CNetwork::getTCPData(net->m_TCPSocket,&buf);
      if (res < 0)
      {
	char *errstr=SDLNet_GetError();
	throw CException("getTCPData: %s\n",
			 strlen(errstr)?errstr:"disconnected");
      }
      else if (res == 0)
      {
	// WARNING (FIXME) mutex ?
	net->print("CNetwork_TCP_client_thread: Disconnected from server !\n");
	SDLNet_TCP_Close(net->m_TCPSocket);
	net->m_TCPSocket=NULL;
	net->m_bClientStarted=false;
	net->m_bReady=false;
	net->m_bError=true;
	break;
      }
      else
      {
	// message
	net->extractMessage(NULL,NULL,buf,res);
      }
    }
    if (numready != 0)
      cerr << "WARNING: numready > 0 (" << numready << ")" << endl;
  }
  // kill the thread
  SDL_KillThread(net->m_TCPThread);
  net->m_TCPThread=NULL;
  return (0);
}
//-----------------------------------------------------------------------
int CNetwork_UDP_thread(void *data)
{
  CNetwork         *net=(CNetwork *)data;
  unsigned char    *buf=NULL;
  SDLNet_SocketSet  sset=NULL;
  int               numready;

  net->printLog("CNetwork_UDP_thread started.\n");
  // prepare the socket set
  sset=SDLNet_AllocSocketSet(1);
  if (sset == NULL)
    throw CException("SDLNet_AllocSocketSet: %s\n", SDLNet_GetError());
  if (SDLNet_UDP_AddSocket(sset,net->m_UDPSocket) == -1)
    throw CException("SDLNet_TCP_AddSocket: %s\n",SDLNet_GetError());
  while (!net->threadShouldStop())
  {
    // listen incoming data
    numready=SDLNet_CheckSockets(sset, (unsigned int)-1);
    // error ?
    if (numready == -1)
      throw CException("SDLNet_CheckSockets: %s\n",SDLNet_GetError());
    // nobody ... strange
    if (numready == 0)
      continue;
    if (SDLNet_SocketReady(net->m_UDPSocket))
    {
      int res;

      res=CNetwork::getUDPData(net->m_UDPSocket,&buf);
      if (res <= 0)
      {
	throw CException("getUDPData: %s\n",SDLNet_GetError());
      }
      else
      {
	IPaddress *ip;
	// message
	ip=SDLNet_UDP_GetPeerAddress(net->m_UDPSocket,-1);
	net->extractMessage(NULL,ip,buf,res);
      }
    }
  }
  SDL_KillThread(net->m_UDPThread);
  net->m_UDPThread=NULL;
  return (0);
}
//-----------------------------------------------------------------------
//***********************************************************************
//                         TCP Raw get/set data
//***********************************************************************
//-----------------------------------------------------------------------
int CNetwork::getTCPData(TCPsocket sock,unsigned char **buf)
{
  static unsigned char *sBuf=NULL; // no mutex because only 
  // one thread acces this !
  static int            sBufLen=0;
  Uint32 len=0,result=0;

  // receive the length of the string message
  result=SDLNet_TCP_Recv(sock,&len,sizeof(len));
  if (result == 0)
    return (0);
  else if (result < sizeof(len))
  {
    if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank!
      throw CException("SDLNet_TCP_Recv: %s\n", SDLNet_GetError());
    return (-1);
  }
  // swap byte order to our local order
  len=SDL_SwapBE32(len);
  // check if anything is strange, like a zero length buffer
  if (len == 0)
    return (0);
  // allocate the buffer memory
  CNetwork::allocBuffer(len,sBuf,sBufLen);
  // get the buffer over the socket
  result=SDLNet_TCP_Recv(sock,sBuf,len);
  if (result < len)
  {
    if(SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank!
      throw CException("SDLNet_TCP_Recv: %s\n", SDLNet_GetError());
    return (-1);
  }
  // give a pointer on data
  *buf=sBuf;
  // return the length read
  return (len);
}
//-----------------------------------------------------------------------
int CNetwork::sendTCPData(TCPsocket sock, unsigned char *buf, int len)
{
  Uint32 result;

  if (buf == NULL || len == 0)
    return (0);

  // change endianness to network order
  len=SDL_SwapBE32(len);

  // send the length of the string
  result=SDLNet_TCP_Send(sock,&len,sizeof(len));
  if (result < sizeof(len)) 
  {
    if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank!
      throw CException("SDLNet_TCP_Send: %s\n", SDLNet_GetError());
  }

  // revert to our local byte order
  len=SDL_SwapBE32(len);

  // send the buffer, with the NULL as well
  result=SDLNet_TCP_Send(sock,buf,len);
  if (result < (Uint32)len) 
  {
    if(SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank!
      throw CException("SDLNet_TCP_Send: %s\n", SDLNet_GetError());
    return (-1);
  }

  // return the length sent
  return (result);
}
//-----------------------------------------------------------------------
//***********************************************************************
//                         UDP Raw get/set data
//***********************************************************************
//-----------------------------------------------------------------------
int CNetwork::getUDPData(UDPsocket sock,unsigned char **buf)
{
  static unsigned char *sBuf=NULL; // no mutex because only 
  // one thread acces this !
  static int            sBufLen=0;
  Uint32                result=0;
  static UDPpacket     *packet=NULL;

  //  std::cerr << "getUDPData - ";
  // alloc packet at first call
  if (packet == NULL)
  {
    packet=SDLNet_AllocPacket(CNETWORK_UDPPACKET_SIZE);
    if (packet == NULL)
      throw CException("CNetwork::getUDPData - Cannot allocate packet");
  }
  SDLNet_UDP_Recv(sock,packet);
  if (packet->len <= 0)
    return (0);
  Uint32 host=SDL_SwapBE32(packet->address.host);
  /*
    std::cerr << " IP Address " <<
    (host>>24) << '.' <<
    ((host>>16) & 0xff) << '.' <<
    ((host>>8) & 0xff) << '.' <<
    (host & 0xff);
    std::cerr << " Port " << SDL_SwapBE16(packet->address.port) << endl;
  */
  // allocate the buffer memory
  CNetwork::allocBuffer(packet->len,sBuf,sBufLen);
  // copy packet content to buffer
  memcpy(sBuf,packet->data,packet->len);
  // give a pointer on data
  *buf=sBuf;
  // return length
  return (packet->len);
}
//-----------------------------------------------------------------------
int CNetwork::sendUDPData(UDPsocket sock,const IPaddress& ip,
						  unsigned char *buf, int len)
{
  static UDPpacket *packet=NULL;

  //  std::cerr << "sendUDPData - ";
  // alloc packet at first call
  if (packet == NULL)
  {
    packet=SDLNet_AllocPacket(CNETWORK_UDPPACKET_SIZE);
    if (packet == NULL)
      throw CException("CNetwork::sendUDPData - Cannot allocate packet");
  }
  if (buf == NULL || len == 0)
    return (0);
  // prepare packet
  if (len > packet->maxlen)
    return (-1);
  packet->len=len;
  memcpy(packet->data,buf,len);
  packet->address=ip;
  packet->channel=-1;
  /*
    Uint32 host=SDL_SwapBE32(packet->address.host);
    std:cerr << " IP Address " <<
    (host>>24) << '.' <<
    ((host>>16) & 0xff) << '.' <<
    ((host>>8) & 0xff) << '.' <<
    (host & 0xff);
    std:cerr << " Port " << SDL_SwapBE16(packet->address.port) << endl;
  */
  // send the packet
  if (!SDLNet_UDP_Send(sock,-1,packet)) 
  {
    if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank!
      throw CException("SDLNet_UDP_Send: %s\n", SDLNet_GetError());
  }
  // return the length sent
  return (len); // FIXME perhaps return (packet->status); ??
}

//-----------------------------------------------------------------------
void CNetwork::startServerThreads()
{
  m_TCPThread=SDL_CreateThread(CNetwork_TCP_server_thread,this);
  if(!m_TCPThread)
  {
    throw CException("SDL_CreateThread: %s\n",SDL_GetError());
  }
  m_UDPThread=SDL_CreateThread(CNetwork_UDP_thread,this);  
  if(!m_UDPThread)
  {
    throw CException("SDL_CreateThread: %s\n",SDL_GetError());
  }
}
//-----------------------------------------------------------------------
void CNetwork::startClientThreads()
{
  m_TCPThread=SDL_CreateThread(CNetwork_TCP_client_thread,this);
  if(!m_TCPThread)
  {
    throw CException("SDL_CreateThread: %s\n",SDL_GetError());
  }
  m_UDPThread=SDL_CreateThread(CNetwork_UDP_thread,this);  
  if(!m_UDPThread)
  {
    throw CException("SDL_CreateThread: %s\n",SDL_GetError());
  }
}

//-----------------------------------------------------------------------
// ---------------------------------------------------------- 
// Sylvain Lefebvre - (c) 2002 - Sylvain.Lefebvre@laposte.net
// ----------------------------------------------------------
