// -------------------------------------------------------------------------
// CVertex class - about a vertex
//
// History: 10/03/2001 ... creation
//
// Author : Sylvain Lefebvre - sylvain.lefebvre@imag.fr 
//                           - http://www.aracknea.net
// -------------------------------------------------------------------------
#ifndef	__CVERTEX_H_
#define	__CVERTEX_H_
// -------------------------------------------------------------------------
#include "CException.h"
// -------------------------------------------------------------------------
#define BIPRODUCT(a,b,c) (( ((b)-(a)).vect((c)-(a)) ).norme())
// -------------------------------------------------------------------------
class	CVertex
{
 public:
 protected:
  static double s_dEpsilon;
  double      m_dX, m_dY, m_dZ;
 public:
  CVertex();
  CVertex(const double& x_, const double& y_, const double& z_);
  CVertex(const double& x_, const double& y_);

  double		     dist(const CVertex& p) const;
  inline double	     length() const;
  inline double	     sqrdLength() const;
  inline CVertex	 unit() {CVertex tmp(m_dX,m_dY,m_dZ); tmp.normalize(); return (tmp);}
  double                 normalize();
  static inline double   getEpsilon();
  static inline double   getEpsilon2() {return (getEpsilon()*getEpsilon());}
  static inline void     setEpsilon(double e);
  inline CVertex	     normMult(const double m[4][4]);
  inline CVertex	     vect(const CVertex & p) const;
  inline CVertex	     cross(const CVertex & p) const;
  inline double	         dot(const CVertex & p) const;
  inline CVertex         mul(const CVertex & p) const;
  inline void            axis(CVertex& u,CVertex& v) const;
  inline void            gl() const;

  inline double x() const {return (m_dX);}
  inline double y() const {return (m_dY);}
  inline double z() const {return (m_dZ);}
  inline void   setX(double d){m_dX=d;}
  inline void   setY(double d){m_dY=d;}
  inline void   setZ(double d){m_dZ=d;}
  inline void   set(double x,double y,double z){m_dX=x;m_dY=y;m_dZ=z;}
  inline double& operator[](int i);
  inline const double& operator[](int i) const;

  inline bool	     equal(const CVertex & a) const;
  inline bool operator== (const CVertex&) const;
  inline bool operator!= (const CVertex&) const;
 
  inline CVertex& operator+= (const CVertex& rkVector);
  inline CVertex& operator-= (const CVertex& rkVector);
  inline CVertex& operator*= (double);
  inline CVertex& operator/= (double);

  inline CVertex operator+ (const CVertex&) const;
  inline CVertex operator- (const CVertex&) const;
  inline CVertex operator* (double) const;
  inline CVertex operator/ (double) const;
  inline CVertex operator- () const;
  inline friend CVertex operator* (double, const CVertex&);

};
// -------------------------------------------------------------------------
#include "CVertex.inl"
// -------------------------------------------------------------------------
#endif /*__CVERTEX_H_*/
// -------------------------------------------------------------------------
